@@ -1,20 +1,18 @@
-use anyhow::{anyhow, Context, Result};
-use clap::{App, AppSettings, Arg, ArgMatches};
-use clap_generate::{generate, generators};
-
-use std::ffi::OsStr;
 use std::io;
 use std::path::Path;
 use std::process::Command;
 
+use anyhow::{anyhow, Context, Result};
+use clap::{App, AppSettings, Arg, ArgMatches};
+use clap_generate::{generate, Shell};
+
 mod kbs2;
 
-fn app<'a, P: AsRef<OsStr>>(default_config_dir: &'a P, default_store_dir: &'a P) -> App<'a> {
+fn app() -> App<'static> {
     // TODO(ww): Put this in a separate file, or switch to YAML.
     // The latter probably won't work with env!, though.
     App::new(env!("CARGO_PKG_NAME"))
         .setting(AppSettings::AllowExternalSubcommands)
-        .setting(AppSettings::VersionlessSubcommands)
         .version(env!("CARGO_PKG_VERSION"))
         .about(env!("CARGO_PKG_DESCRIPTION"))
         .arg(
@@ -25,7 +23,7 @@ fn app<'a, P: AsRef<OsStr>>(default_config_dir: &'a P, default_store_dir: &'a P)
                 .value_name("DIR")
                 .takes_value(true)
                 .env("KBS2_CONFIG_DIR")
-                .default_value_os(default_config_dir.as_ref()),
+                .default_value_os(kbs2::config::DEFAULT_CONFIG_DIR.as_ref()),
         )
         .arg(
             Arg::new("completions")
@@ -33,7 +31,7 @@ fn app<'a, P: AsRef<OsStr>>(default_config_dir: &'a P, default_store_dir: &'a P)
                 .long("completions")
                 .value_name("SHELL")
                 .takes_value(true)
-                .possible_values(&["bash", "zsh", "fish"]),
+                .possible_values(Shell::arg_values()),
         )
         .subcommand(
             App::new("agent")
@@ -54,6 +52,10 @@ fn app<'a, P: AsRef<OsStr>>(default_config_dir: &'a P, default_store_dir: &'a P)
                                 .long("quit"),
                         ),
                 )
+                .subcommand(
+                    App::new("query")
+                        .about("ask the current agent whether it has the current config's key"),
+                )
                 .subcommand(
                     App::new("unwrap")
                         .about("unwrap the current config's key in the running agent"),
@@ -75,7 +77,7 @@ fn app<'a, P: AsRef<OsStr>>(default_config_dir: &'a P, default_store_dir: &'a P)
                         .long("store-dir")
                         .value_name("DIR")
                         .takes_value(true)
-                        .default_value_os(default_store_dir.as_ref()),
+                        .default_value_os(kbs2::config::DEFAULT_STORE_DIR.as_ref()),
                 )
                 .arg(
                     Arg::new("insecure-not-wrapped")
@@ -147,25 +149,27 @@ fn app<'a, P: AsRef<OsStr>>(default_config_dir: &'a P, default_store_dir: &'a P)
                 ),
         )
         .subcommand(
-            App::new("rm").about("remove a record").arg(
+            App::new("rm").about("remove one or more records").arg(
                 Arg::new("label")
-                    .about("the record's label")
+                    .about("the labels of the records to remove")
                     .index(1)
-                    .required(true),
+                    .required(true)
+                    .multiple_values(true),
             ),
         )
         .subcommand(
             App::new("dump")
-                .about("dump a record")
+                .about("dump one or more records")
                 .arg(
                     Arg::new("label")
-                        .about("the record's label")
+                        .about("the labels of the records to dump")
                         .index(1)
-                        .required(true),
+                        .required(true)
+                        .multiple_values(true),
                 )
                 .arg(
                     Arg::new("json")
-                        .about("dump in JSON format")
+                        .about("dump in JSON format (JSONL when multiple)")
                         .short('j')
                         .long("json"),
                 ),
@@ -272,7 +276,7 @@ fn run(matches: &ArgMatches, config: &kbs2::config::Config) -> Result<()> {
 
     // Special case: `kbs2 agent` does not receive pre- or post-hooks.
     if let Some(("agent", matches)) = matches.subcommand() {
-        return kbs2::command::agent(&matches, &config);
+        return kbs2::command::agent(matches, config);
     }
 
     if let Some(pre_hook) = &config.pre_hook {
@@ -281,16 +285,16 @@ fn run(matches: &ArgMatches, config: &kbs2::config::Config) -> Result<()> {
     }
 
     match matches.subcommand() {
-        Some(("new", matches)) => kbs2::command::new(&matches, &config)?,
-        Some(("list", matches)) => kbs2::command::list(&matches, &config)?,
-        Some(("rm", matches)) => kbs2::command::rm(&matches, &config)?,
-        Some(("dump", matches)) => kbs2::command::dump(&matches, &config)?,
-        Some(("pass", matches)) => kbs2::command::pass(&matches, &config)?,
-        Some(("env", matches)) => kbs2::command::env(&matches, &config)?,
-        Some(("edit", matches)) => kbs2::command::edit(&matches, &config)?,
-        Some(("generate", matches)) => kbs2::command::generate(&matches, &config)?,
-        Some(("rewrap", matches)) => kbs2::command::rewrap(&matches, &config)?,
-        Some(("rekey", matches)) => kbs2::command::rekey(&matches, &config)?,
+        Some(("new", matches)) => kbs2::command::new(matches, config)?,
+        Some(("list", matches)) => kbs2::command::list(matches, config)?,
+        Some(("rm", matches)) => kbs2::command::rm(matches, config)?,
+        Some(("dump", matches)) => kbs2::command::dump(matches, config)?,
+        Some(("pass", matches)) => kbs2::command::pass(matches, config)?,
+        Some(("env", matches)) => kbs2::command::env(matches, config)?,
+        Some(("edit", matches)) => kbs2::command::edit(matches, config)?,
+        Some(("generate", matches)) => kbs2::command::generate(matches, config)?,
+        Some(("rewrap", matches)) => kbs2::command::rewrap(matches, config)?,
+        Some(("rekey", matches)) => kbs2::command::rekey(matches, config)?,
         Some((cmd, matches)) => {
             let cmd = format!("kbs2-{}", cmd);
 
@@ -333,27 +337,13 @@ fn run(matches: &ArgMatches, config: &kbs2::config::Config) -> Result<()> {
 fn main() -> Result<()> {
     env_logger::init();
 
-    let default_config_dir = kbs2::config::find_default_config_dir()?;
-    let default_store_dir = kbs2::config::find_default_store_dir()?;
-
-    let mut app = app(&default_config_dir, &default_store_dir);
+    let mut app = app();
     let matches = app.clone().get_matches();
 
     // Shell completion generation is completely independent, so perform it before
     // any config or subcommand operations.
-    if let Some(shell) = matches.value_of("completions") {
-        match shell {
-            "bash" => {
-                generate::<generators::Bash, _>(&mut app, env!("CARGO_PKG_NAME"), &mut io::stdout())
-            }
-            "zsh" => {
-                generate::<generators::Zsh, _>(&mut app, env!("CARGO_PKG_NAME"), &mut io::stdout())
-            }
-            "fish" => {
-                generate::<generators::Fish, _>(&mut app, env!("CARGO_PKG_NAME"), &mut io::stdout())
-            }
-            _ => unreachable!(),
-        }
+    if let Ok(shell) = matches.value_of_t::<Shell>("completions") {
+        generate(shell, &mut app, env!("CARGO_PKG_NAME"), &mut io::stdout());
         return Ok(());
     }
 
@@ -372,7 +362,7 @@ fn main() -> Result<()> {
             .print_long_help()
             .with_context(|| "failed to print help".to_string());
     } else if let Some(("init", matches)) = matches.subcommand() {
-        return kbs2::command::init(&matches, &config_dir);
+        return kbs2::command::init(matches, config_dir);
     }
 
     // Everything else (i.e., all other subcommands) go through here.
